unit ListDir;

interface

uses
  sysutils, classes, contnrs, Dialogs, comctrls,
  FTPSend, md5;


  type
  TFTPError = (feConnected, feNotConnected, feDirNotFound, feCouldNotCreate);
  TLocalErrorType = (deOK, deFoundDir);
  TListEvent = procedure(count: integer; filename: AnsiString) of object;
  TProgEvent = procedure(count: integer; filename: AnsiString) of object;
  TErrorEvent = procedure(value: AnsiString) of object;
  
  TLocalError = class(Tobject)
    private
      FError: TLocalErrorType;
      FDir: AnsiString;
    public
      property Error: TLocalErrorType read FError write FError;
      property Dir: AnsiString read FDir write FDir;
  end;
  { TDirList }

  { TMyStringList }

  TMyStringList = class(TStringList)
    private
      FDelimitedText: string;
      procedure SetDelimitedText(const AValue: string);
    public
      property DelimitedText: string read FDelimitedText write SetDelimitedText;
  end;

  TDirList = class(TObjectList)
    private
    FSorted: boolean;
    public
    procedure sort; reintroduce;
    procedure LoadFromFile(FileName: string);
    function GetStrings(booleans: boolean): TStrings;
    procedure SaveToFile(FileName: string);
  end;

  { TListDir }
  
  TStrArray = array of AnsiString;

  TListDir = class(TObject)
    private
      FProgressEvent: TProgEvent;
      FCompareEvent: TListEvent;
      FLocalListEvent: TListEvent;
      FFTPListEvent: TListEvent;
      FErrorEvent: TErrorEvent;
      FDirectoriesLocal: integer;
      FDirectoriesFTP: integer;
      FCompared: integer;
      FLoadFiles: integer;
      FDelFiles: integer;
      FUserName: string;
      FPassWord: string;
      FHost: string;
      FPort: string;
      FLogFileName: string;
      FResultList: TDirList;
      FAbort: Boolean;
      FIgnore: Boolean;
      FWriteLog: Boolean;
      Fmd5FTP: boolean;
      FPassive: boolean;
      DateBase: TDateTime;
      HomeList: TDirList;
      LocalList: TDirList;
      FTPList: TDirList;
      FIgnoreArray: TStrArray;
      FFTPIgnoreArray: TStrArray;
      LocalDirBase: string;
      FTPDirBase: string;
      AFTPsend: TFTPSend;
      OldDir: string;
      function ListDirFiles(DirName: string; name : string): TLocalError;
      procedure ListFTPFiles(name: string);
      function GetFilesLocal: integer;
      function GetFilesFTP: integer;
      procedure SetPassive(const AValue: boolean);

    public
      property DirectoriesLocal: integer read FDirectoriesLocal;
      property FilesLocal: integer read GetFilesLocal;
      property DirectoriesFTP: integer read FDirectoriesFTP;
      property FilesFTP: integer read GetFilesFTP;
      property Compared: integer read FCompared;
      property LoadFiles: integer read FLoadFiles;
      property DelFiles: integer read FDelFiles;
      property ResultList: TDirList read FResultList;
      function CallFTPFiles(DirName: string; list: TStrings; CreateDir: boolean): TFTPError;
      function CallDirFiles(DirName: string; list: TStrings): boolean;
      procedure CompareDirFiles;
      procedure Commit;
      property UserName: string read FUserName write FUserName;
      property PassWord: string read FPassWord write FPassWord;
      property Host: string read FHost write FHost;
      property Port: string read FPort write FPort;
      property LogFileName: string read FLogFileName write FLogFileName;
      property ProgressEvent: TProgEvent read FProgressEvent write FProgressEvent;
      property LocalListEvent: TListEvent read FLocalListEvent write FLocalListEvent;
      property FTPListEvent: TListEvent read FFTPListEvent write FFTPListEvent;
      property CompareEvent: TListEvent read FCompareEvent write FCompareEvent;
      property ErrorEvent: TErrorEvent read FErrorEvent write FErrorEvent;
      property Abort: Boolean read FAbort write FAbort;
      property Ignore: Boolean read FIgnore write FIgnore;
      property WriteLog: Boolean read FWriteLog write FWriteLog;
      property IgnoreArray: TStrArray read FIgnoreArray write FIgnoreArray;
      property FTPIgnoreArray: TStrArray read FFTPIgnoreArray write FFTPIgnoreArray;
      property md5FTP: boolean read Fmd5FTP write Fmd5FTP;
      property Passive: boolean read FPassive write SetPassive;
      constructor create;
  end;
  
  
  TLoad = record
    load, overwrite: boolean;
  end;
  
  TFileListItem = class(Tobject)
  private
    FName : string;
    FSize : string;
    FTime : string;
    FMD5  : string;
    FUpLoad : TLoad;
    FDel  : boolean;
    FFTPTime : TDateTime;
    FLOcalTime : TDateTime;
    NewList : boolean;
  published
    { Private declarations }
  public
    property Name : string read FName write FName;
    property Size : string read FSize write FSize;
    property Time : string read FTime write FTime;
    property LocalTime : TDateTime read FLocalTime write FLocalTime;
    property FTPTime : TDateTime read FFTPTime write FFTPTime;
    property MD5  : string read FMD5 write FMD5;
    property UpLoad : TLoad read FUpLoad write FUpLoad;
    property Del  : boolean read FDel write FDel;

    constructor create;
    { Public declarations }
  end;


implementation



{var
  DateBase: TDateTime;
  HomeList: TDirList;
  LocalList: TDirList;
  FTPList: TDirList;
  LocalDirBase: string;
  FTPDirBase: string;
  ResultList: TDirList;
  AFTPsend: TFTPSend;    }

const
  delimiter= ':';
  datebasestr= '~~datebase~~';
  dirbasestr= '~~DIR~~';
  direrrorstr1= 'Found "'+dirbasestr+'" in "';
  direrrorstr2= '" please remove this file from the compared directories';
  {TListDir}
  
function strisin(value: string; arr: TStrArray): boolean;
var
  i: integer;
begin
   result:= false;
  if value[1] = '/' then value:= copy(value, 2, high(integer));
  if value[length(value)] = '/' then value:= copy(value, 1, length(value)-1);
  for i:= 0 to high(arr) do begin
    if value = arr[i] then begin
      result:= true;
      break;
    end;
  end;
end;

function TListDir.CallDirFiles(DirName: string; list: TStrings): boolean;
var
  error: TLocalError;
begin
  error:= TLocalError.Create;
  FDirectoriesLocal:= 0;
  LocalDirBase:= DirName;

  LocalList.clear;
  error:= listdirfiles(DirName, '');
  if error.error = deFoundDir then begin
    if assigned(ErrorEvent) then ErrorEvent(direrrorstr1 + error.Dir + direrrorstr2);
    result:= false;
  end;
  list.Clear;
  list.AddStrings(LocalList.GetStrings(false));
end;

function TListDir.CallFTPFiles(DirName: string; list: TStrings; CreateDir: boolean): TFTPError;
var
  i: integer;
begin
  FDirectoriesFTP:= 0;
  AFTPSend.Timeout:= 10000;
  AFTPSend.DirectFile:= false;
  AFTPsend.UserName:= FUserName;
  AFTPsend.Password:= FPassWord;
  AFTPsend.TargetHost:= FHost;
  AFTPsend.TargetPort:= FPort;
  

  FTPList.Clear;
  if AFTPsend.Login then begin
    FTPDirBase:= DirName;
    if not AFTPsend.ChangeWorkingDir(DirName) then begin
      if CreateDir then begin
        if AFTPsend.CreateDir(DirName)then begin
          if not AFTPsend.ChangeWorkingDir(DirName) then begin
            result:= feCouldNotCreate;
            exit;
          end;
         end else begin
          result:= feCouldNotCreate;
          exit;
        end;
       end else begin
        result:= feDirNotFound;
        exit;
      end;
    end;
      AFTPsend.List(AFTPsend.GetCurrentDir, false);
      for I := 0 to AFTPsend.FtpList.Count-1 do begin
         if AFTPsend.FtpList.Items[i].FileName = datebasestr then begin
           DateBase:= AFTPsend.FtpList.Items[i].FileTime;
           break;
         end;
      end;
      ListFTPFiles('');
      list.Clear;
      list.AddStrings(FTPList.GetStrings(false));
    AFTPsend.ChangeToParentDir;
    Result:= feConnected;
    AFTPsend.Logout;
  end else begin
    Result:= feNotConnected;
  end;
end;

function TListDir.ListDirFiles(DirName: string; name : string): TLocalError;
var
  rec: TSearchRec;
  i: integer;
  n: string;
  d: tdatetime;
  item: TFileListItem;
begin
  result:= TLocalError.create;
  result.Error:= deOK;
  result.Dir:= '';
  inc(FDirectoriesLocal);
  DirName:= IncludeTrailingPathDelimiter(DirName);
  i:= FindFirst(DirName+'*', faAnyFile, rec);
  if assigned(FLocalListEvent) then FLocalListEvent(FDirectoriesLocal, DirName);
  while i = 0 do begin
    if not FAbort then begin
      if (rec.Name <> dirbasestr) or (name = '')then begin
        n:= name + '/' + rec.name;
        if ((rec.Attr and faDirectory) = 0) then begin
          if (rec.name <> dirbasestr) then begin
            item:= TFileListItem.create;
            d:= FileDateToDateTime(rec.Time) * SecsPerDay;
   //       list.Add(n + delimiter + IntToStr(rec.Size) + delimiter + IntToStr(trunc(d)));// + delimiter + MD5Print(MD5File(Name + rec.Name)));
            item.name:=n;
            item.Size:= IntToStr(rec.Size);
            item.Time:= IntToStr(trunc(d));
            item.LocalTime:= FileDateToDateTime(rec.Time);
            LocalList.Add(item);
          end;
         end else begin
          if ignore then begin
            if (rec.Name <> '.') and (rec.Name <> '..') and (not strisin(rec.name, IgnoreArray)) then begin
              result:= ListDirFiles(DirName+rec.Name, n);
              if (result.error = deFoundDir) and (result.Dir = '') then begin
                result.Dir:= name + '/' + rec.name;
                break;
              end;
            end;
           end else begin
            if (rec.Name <> '.') and (rec.Name <> '..') then begin
              result:= ListDirFiles(DirName+rec.Name, n);
              if (result.error = deFoundDir) and (result.Dir = '') then begin
                result.Dir:= name + '/' + rec.name;
                break;
              end;
            end;
          end;
        end;
        i:= FindNext(rec);
      end else begin
        result.error:= deFoundDir;
        break;
      end;
   end else break;
  end;
  FindClose(rec);
end;

procedure TListDir.ListFTPFiles(name: string);
var
  i, listcount: integer;
  s, n: string;
  aftplist: TFTPList;
  item: TFileListItem;
begin
  inc(FDirectoriesFTP);
  if assigned(FFTPListEvent) then FFTPListEvent(FDirectoriesFTP, name);
  AFTPSend.list(AFTPSend.GetCurrentDir, false);
  aftplist:= TFTPList.Create;
  aftplist.Assign(AFTPSend.FtpList);       // Warum ist Assign notwendig ?!?!?!
//  aftplist:=AFTPSend.FtpList ;
  if assigned(aftplist) then begin
    listcount:=aftpList.Count;
   end else begin
    listcount:=0;
  end;
  for I := 0 to listcount-1 do begin
    if not FAbort then begin
      n := aftpList.Items[i].FileName;
      n:= name + '/' + n;
      if not aftpList.Items[i].Directory then begin
 {  s:= n + delimiter;
    s:= s + IntToStr(FtpList.Items[i].FileSize) + delimiter;
    s:= s + IntToStr(trunc((FtpList.Items[i].FileTime - DateBase) * SecsPerDay));
 }
        item:= TFileListItem.create;
        item.Name:= n;
        item.Size:= IntToStr(aftpList.Items[i].FileSize);
        item.Time:= IntToStr(trunc((aftpList.Items[i].FileTime - DateBase) * SecsPerDay));
        item.FTPTime:= aftpList.Items[i].FileTime;
        FTPList.add(item);
       end else begin
        if AFTPSend.ChangeWorkingDir(aftpList.Items[i].FileName) then begin
          if not(strisin(n, FTPIgnoreArray)) then begin
            ListFTPFiles(n);
          end;
          if not AFTPSend.ChangeToParentDir then
            FAbort:= true;
        end;
      end;
     end else begin
      break;
    end;
  end;
  aftplist.Free;
end;


function TListDir.GetFilesFTP: integer;
begin
  result:= FTPList.count;
end;

procedure TListDir.SetPassive(const AValue: boolean);
begin
  AFTPsend.PassiveMode:= AValue;
end;

function TListDir.GetFilesLocal: integer;
begin
  result:= LocalList.count;
end;

procedure TListDir.CompareDirFiles;
var
  h, l ,f, fold: integer;
  tmphList,tmplList,tmpfList: TStringlist;
  home, local, ftp, ftpold: string;
  chl, chf, clf, cc: Integer;
  dir: string;
  b: Boolean;
const
  HighChar = #$FF;
  HighString = HighChar + delimiter + '0'+ delimiter + '0'+ delimiter + '0';

  function cadjust(c: integer) : integer;
  begin
    if c<0 then begin
      c := -1;
     end else begin
      if c>0 then c := 1;
    end;
    Result := c;
  end;

  procedure WriteList(h,l,f: integer);
  var
    fi: TFileListItem;
    cc : integer;
    ftemp : TStringList;
    newmd5,name: String;
    i: integer;
    dl: TFileStream;
    s: String;

  begin

//    if h = 856 then
//      beep;


    if h>=0 then cc := 1 else cc := 0;
    if l>=0 then cc := cc+2;
    if f>=0 then cc := cc+4;
    fi :=  TFileListItem.create;
    case cc of
      0: begin  // unmglich
        showmessage('BLDSINN version 2!');
      end;
      1, 5: begin         // h aber nicht l
//        ftemp := tmphlist;
        fi.Name := (HomeList[h]as TFileListItem).Name;
        fi.Size := (HomeList[h]as TFileListItem).Size;
        fi.time := (HomeList[h]as TFileListItem).Time;
        if f >= 0 then
          fi.FTPTime := (FTPList[f]as TFileListItem).FTPTime;
      end;
      2, 3, 6, 7: begin  // mindestens l
//        ftemp := tmpllist;
        fi.Name := (LocalList[l]as TFileListItem).Name;
        fi.Size := (LocalList[l]as TFileListItem).Size;
        fi.time := (LocalList[l]as TFileListItem).Time;
        fi.LocalTime:= (LocalList[l]as TFileListItem).LocalTime;
        if f >= 0 then
          fi.FTPTime := (FTPList[f]as TFileListItem).FTPTime;
      end;
      4: begin  // nur f
//        ftemp := tmpflist;
        fi.Name := (FTPList[f]as TFileListItem).Name;
        fi.Size := (FTPList[f]as TFileListItem).Size;
        fi.time := (FTPList[f]as TFileListItem).Time;
        fi.FTPTime := (FTPList[f]as TFileListItem).FTPTime;
      end;
    end;
//    fi.Name := ftemp[0];
//    fi.Size := ftemp[1];
//    fi.Time := ftemp[2];
    case cc  of
      0, 1: begin        // nur h
        fi.Free;
        fi:= NIL;
      end;
      2: begin        // nur l
        fi.UpLoad.load:= true;
        fi.UpLoad.overwrite:= false;
        inc(FLoadFiles);
        fi.MD5:= MD5Print(MD5File(IncludeTrailingPathDelimiter(LocalDirBase) + fi.Name));
      end;
      6: begin   // l und f
        fi.NewList:= true;
        if Fmd5FTP then begin
//          if AFTPsend.Login then begin
//            AFTPsend.justmd5:= true;
//            if AFTPsend.RetrieveFile(copy(FTPDirBase, 3, length(FTPDirBase))+ fi.name, false) or          // ????
//               AFTPsend.RetrieveFile(FTPDirBase+ fi.name, false)
//            then begin
//            s := copy(FTPDirBase, 3, length(FTPDirBase))+ fi.name;
//            b := AFTPsend.RetrieveFile(s, false);
//            if not b then begin
            s := FTPDirBase+ fi.name;
            b := AFTPsend.RetrieveFile(s, false);
//            end;
            if b then begin
              fi.MD5:= AFTPsend.md5;
              newmd5:= MD5Print(MD5File(IncludeTrailingPathDelimiter(LocalDirBase) + fi.Name));
              if fi.MD5 <> newmd5 then begin
                fi.MD5:= newmd5;
                fi.UpLoad.load:= false;
                fi.UpLoad.overwrite:= true;
                Inc(LoadFiles);
               end else begin
                fi.UpLoad.load:= false;
                fi.UpLoad.overwrite:= false;
              end;
             end else begin
              fi.UpLoad.load:= false;
              fi.UpLoad.overwrite:= true;
              inc(FLoadFiles);
              fi.MD5:= MD5Print(MD5File(IncludeTrailingPathDelimiter(LocalDirBase) + fi.Name));
              AFTPsend.Login;
            end;
//           AFTPsend.justmd5:= false;
//           end else begin
//            fi.UpLoad.load:= false;
//            fi.UpLoad.overwrite:= true;
//            inc(FLoadFiles);
//            fi.MD5:= MD5Print(MD5File(IncludeTrailingPathDelimiter(LocalDirBase) + fi.Name));
//          end;
         end else begin
          fi.UpLoad.load:= false;
          fi.UpLoad.overwrite:= true;
          inc(FLoadFiles);
          fi.MD5:= MD5Print(MD5File(IncludeTrailingPathDelimiter(LocalDirBase) + fi.Name));
        end;
      end;
      3: begin       // h und l
        fi.NewList:= true;
        fi.UpLoad.load:= true;
        fi.UpLoad.overwrite:= false;;
        inc(FLoadFiles);
        if (fi.Size <> (HomeList[h]as TFileListItem).Size) or (fi.Time <> (HomeList[h]as TFileListItem).Time) then begin
          newmd5 := MD5Print(MD5File(IncludeTrailingPathDelimiter(LocalDirBase) + fi.Name));
          fi.MD5 := newmd5;
         end else begin
          fi.md5 := (HomeList[h]as TFileListItem).MD5;
        end;
      end;
      4, 5 : begin  // nur f oder h und f
        fi.Del := true;
        inc(FDelFiles);
        if cc = 5 then begin
          fi.NewList:= true;
        end;
      end;
      7: begin       // h und l und f
        fi.NewList:= true;
        if fi.size = ((FTPList[f]as TFileListItem).Size) then begin
          if (fi.Size <> (HomeList[h]as TFileListItem).Size) or (fi.Time <> (HomeList[h]as TFileListItem).Time) then begin
            newmd5 := MD5Print(MD5File(IncludeTrailingPathDelimiter(LocalDirBase) + fi.Name));
            fi.MD5 := newmd5;
            if newmd5 <> (HomeList[h]as TFileListItem).MD5 then begin
              fi.UpLoad.load:= false;
              fi.UpLoad.overwrite:= true;;
              inc(FLoadFiles);
            end;
           end else begin
            fi.md5 := (HomeList[h]as TFileListItem).MD5;
            fi.UpLoad.load:= false;
            fi.UpLoad.overwrite:= fi.Size <> (HomeList[h]as TFileListItem).Size;
            if fi.UpLoad.overwrite or fi.UpLoad.load then
              inc(FLoadFiles);
          end;
         end else begin
          newmd5 := MD5Print(MD5File(IncludeTrailingPathDelimiter(LocalDirBase) + fi.Name));
          fi.MD5 := newmd5;
          fi.UpLoad.load:= false;
          fi.UpLoad.overwrite:= true;
          inc(FLoadFiles);
        end;
      end;
    end;
    if assigned (fi) then begin
      ResultList.Add(fi);
      if assigned(FCompareEvent) then FCompareEvent(FCompared, fi.FName);
    end;
  end;

begin
  FDelFiles:= 0;
  FLoadFiles:= 0;
  FCompared:= 0;
  h:= 0;
  l:= 0;
  f:= 0;
  if assigned(ResultList) then ResultList.Free;
  FResultList:= TDirList.Create;
  tmphList:= TStringList.Create;
 // tmphList.Delimiter:= delimiter;
  tmplList:= TStringList.Create;
//  tmplList.Delimiter:= delimiter;
  tmpfList:= TStringList.Create;
//  tmpfList.Delimiter:= delimiter;
  {$ifndef FPC}
//  tmplList.StrictDelimiter:= true;
//  tmplList.StrictDelimiter:= true;
//  tmpfList.StrictDelimiter:= true;
  {$endif}
//  if Assigned(homeList) then HomeList.Free;
//  HomeList:= TDirList.Create;
  HomeList.Clear;
  dir:= IncludeTrailingPathDelimiter(LocalDirBase)+'~~DIR~~';
  if FileExists(dir) then HomeList.LoadFromFile(dir);
//  Homelist.CaseSensitive:= true;
  HomeList.Sort;
//  LocalList.CaseSensitive:= true;
  LocalList.Sort;
//  FTPList.CaseSensitive:= true;
  FTPList.Sort;
//  FTPList.CaseSensitive:= true;
  ftpold:= '';
  fold:=-1;
  
  if Fmd5FTP then begin
    if AFTPsend.Login then begin
      AFTPsend.justmd5:= true;
     end else begin
      Fmd5FTP:=false;
    end;
  end;
  
  while not FAbort do begin
    if h < HomeList.Count then begin
      home:= (HomeList[h]as TFileListItem).name;
//      tmphList.DelimitedText:= HomeList[h];
     end else begin
      home:= HighChar;
//      tmphList.DelimitedText:= HighString;
    end;
    if l < LocalList.Count then begin
      local:= (LocalList[l]as TFileListItem).name;
//      tmplList.DelimitedText:= LocalList[l];
     end else begin
       local:= HighChar;
//      tmplList.DelimitedText:= HighString;
    end;
    if f < FTPList.Count then begin
      ftp:= (FTPList[f]as TFileListItem).name;

// workaround fuer buggy ftpsend komponente (listet files mehrfach)
      if f <> fold then begin
        while (ftp = ftpold) do begin
          inc(f);
          if f < FTPList.Count then begin
           if assigned(FCompareEvent) then FCompareEvent(FCompared, ftp);
            beep;
            sleep (500);
            ftp:= (FTPList[f]as TFileListItem).name;
           end else begin
            ftp:= HighChar;
          end;
        end;
      end;
      ftpold:=ftp;
      fold:=f;
//      tmpfList.DelimitedText:= FTPList[f];
     end else begin
       ftp:= HighChar;
//      tmpfList.DelimitedText:= HighString;
    end;
//    home:= tmphList[0];
//    local:= tmplList[0];
//    ftp:= tmpfList[0];

    chl := AnsiCompareStr(home, local);
    chl := cadjust(chl);
    chf := AnsiCompareStr(home, ftp);
    chf := cadjust(chf);
    clf := AnsiCompareStr(local, ftp);
    clf := cadjust(clf);
    cc := chl + 3*chf + 9*clf;
    case cc of
      1, 3, 9, -1, -3, -9, 
      2, 6, 7, -2, -6,- 7,
      10, -10: 
      begin  // unmglich
        showmessage('BLDSINN!');
      end;
      000: begin //  l=f=h
        if home = HighChar then begin
          break;
        end;
        Writelist(h,l,f);
        inc(h);
        inc(l);
        inc(f);
      end;
      004: begin // l=f<h
        Writelist(-1,l,f);
        inc(l);
        inc(f);
      end;
      -12: begin // l=h<f
        Writelist(h,l,-1);
        inc(h);
        inc(l);
      end;
      008: begin // h=f<l
        Writelist(h,-1,f);
        inc(h);
        inc(f);
      end;
      -13, -4, 5: begin //  h<l und h<f
        Writelist(h,-1,-1);
        inc(h);
      end;      
      -11, -8, -5: begin //  l<h und l<f     
        Writelist(-1,l,-1);
        inc(l);
      end;   
      11, 12, 13: begin //  f<h und f<l     
        Writelist(-1,-1,f);
        inc(f);
      end;                
    end;
    inc(FCompared);
  end;
  AFTPsend.Logout;
  AFTPsend.justmd5:= false;
end;

constructor TListDir.create;
begin
  FLoadFiles:= 0;
  FDelFiles:= 0;
  FDirectoriesLocal:= 0;
  FDirectoriesFTP:= 0;
  AFTPSend:= TFTPSend.Create;
  FTPList:= TDirList.Create;
  HomeList:= TDirList.Create;
  LocalList:= TDirList.Create;
  FResultList:= TDirlist.Create;
end;

procedure TListDir.Commit;
var
  i, j, load, del: integer;
  tmp: TFileListItem;
  name, dir, path, s: AnsiString;
  stream: TFileStream;
  count: integer;
  timeout: integer;
  LogFile: TFileStream;


  function ExtractFTPFilePath(filename: string): string;
  var
  I: Integer;
  begin
    I := LastDelimiter('/' + '', FileName);
    Result := Copy(FileName, 1, I);
  end;
  function ExtractFTPFileName(const FileName: string): string;
  var
    I: Integer;
  begin
    I := LastDelimiter('/' + '', FileName);
    Result := Copy(FileName, I + 1, MaxInt);
  end;
  
  function RemoveFTPDelimiter(FileName: string; LastOnly: boolean ): string;
  var
  I:  integer;
  begin
    if LastOnly then begin
      if Filename[length(FileName)] = '/' then delete(FileName, length(FileName), MaxInt);
     end else begin
      i:= pos('/', filename);
      while i <> 0 do begin
        delete(filename,i,1);
        i:= pos('/', filename);
      end;
    end;
    result:= filename;
  end;
  
  function ReplaceFTPDelimiter(FileName: string): string;
  begin
  {$ifdef unix}
    Result := FileName;
  {$else}
    Result := StringReplace(FileName, '/', '\', [rfReplaceAll]);
  {$endif}
  end;
  
  function ExtractFTPDir(const FileName: string): string;
  var
    I: Integer;
  begin
    I := LastDelimiter('/' + '', FileName);
    Result := Copy(FileName, 1, I-1);
    I := LastDelimiter('/' + '',Result);
    Result := Copy(Result, I + 1, MaxInt);
  end;
  
  function ChangeDir(newdir: String): boolean;
    var
      list: TMyStringList;
      i, n: integer;
      tmps1, tmps2: string;
    begin
      result:= true;
      if newdir = OldDir then exit;
      tmps1 := OldDir;
      list:= TMyStringList.Create;
      list.Delimiter:= '/';
      if length(tmps1) > 0 then begin
        if tmps1[1] = '/' then delete(tmps1, 1, 1);
      end;
      list.DelimitedText:= tmps1;
      OldDir:= newdir;
      tmps1:= '';
      n:= List.Count;
      if n > 0 then begin
        for i:= 0 to n - 1 do begin
          tmps1:=  tmps1 + '../';
        end;
        if not AFTPsend.ChangeWorkingDir(tmps1) then begin
          list.free;
          result:= false;
          exit;
        end;
      end;
      if length(newdir) > 0 then begin
        if newdir[1] = '/' then delete(newdir, 1, 1);
      end;
      list.DelimitedText:= newdir;
      n := List.Count;
      for i := 0 to n - 1 do begin
        tmps1:= list[i];
        tmps2:= AFTPsend.GetCurrentDir;
        if not AFTPsend.ChangeWorkingDir(tmps1) then begin
          AFTPsend.CreateDir(tmps1);
          if not AFTPsend.ChangeWorkingDir(tmps1) then begin
            list.free;
            result:= false;
            exit;
          end;
        end;
      end;
      list.free;
    end;
  
  function doload:boolean;
  var
    temps1, temps2, aktname: String;
  label start;


  begin
    if assigned(stream) then FreeAndNil(stream);
    temps1 := LocalDirBase;
    temps2 := ReplaceFTPDelimiter(name);
    stream:= TFileStream.Create(temps1+temps2, fmOpenRead);
    timeout:= 0;
    start:
    inc(timeout);
    if not changedir(dir) then begin
      if AFTPsend.Login then begin
        if timeout <= 10 then goto start
         else begin
          result:= false;
          exit;
        end;
       end else begin
        result:= false;
        exit;
      end;
    end;
    aktname:= ExtractFTPFileName(tmp.name);
    AFTPsend.DeleteFile(aktname);
    AFTPSend.DataStream.Clear;
    AFTPSend.DataStream.CopyFrom(stream,0);
    if not AFTPsend.StoreFile(aktname, false) then begin
      if AFTPsend.login then begin
        if timeout <= 10 then goto start
         else begin
          result:= false;
          exit;
        end;
       end else begin
        result:= false;
        exit;
      end
    end;
{
    temps1:=FTPDirBase;
    if not AFTPsend.ChangeWorkingDir(temps1) then begin
      if AFTPsend.login then begin
        if timeout <= 10 then goto start
         else begin
          result:= false;
          exit;
        end;
       end else begin
        result:= false;
        exit;
      end
    end;
}
    FreeAndNil(stream);
  end;
  
  function dodelete:boolean;
  var
    temps1: string;
  label start;
  begin
    timeout:= 0;
    start:
    inc(timeout);
    ChangeDir(dir);
    if not AFTPsend.DeleteFile(name) then begin
      if AFTPsend.login then begin
        if timeout <= 10 then goto start
         else begin
          result:= false;
          exit;
        end;
       end else begin
        result:= false;
        exit;
      end
    end;
    dir:= AFTPsend.GetCurrentDir;
    AFTPsend.List(dir , false);
    if AFTPsend.FtpList.Count = 0 then begin
      if AFTPsend.ChangeToParentDir then begin
        temps1:= ExtractFTPFilePath(olddir);
        if temps1[length(temps1)] = '/' then delete(temps1,length(temps1),1);
        olddir:= temps1;
        dir:= ExtractFTPFilename(dir);
        if not AFTPsend.DeleteDir(dir) then begin
          if AFTPsend.login then begin
            if timeout <= 10 then goto start
             else begin
              result:= false;
              exit;
            end;
           end else begin
            result:= false;
            exit;
          end
        end;
       end else begin
        if AFTPsend.login then begin
          if timeout <= 10 then goto start
           else begin
            result:= false;
            exit;
          end;
         end else begin
          result:= false;
          exit;
        end
      end;
    end;
//    AFTPsend.ChangeWorkingDir(FTPDirBase);
  end;
  
begin
  if AFTPsend.Login then begin
    LogFile:= NIL;
    if WriteLog then begin
      if not FileExists(LogFileName) then begin
        LogFile:= TFileStream.Create(LogFileName, fmCreate)
       end else begin
        LogFile:= TFileStream.Create(LogFileName, fmOpenReadWrite);
      end;
      LogFile.Seek(0, soFromEnd);
      s:= DateTimeToStr(now)+' '+ LocalDirBase + ' -> ' +  FHost + ' : ' + FTPDirBase +
          ' copied:' + inttostr(FLoadFiles)  + ' deleted:' + inttostr(FDelFiles)  +
          #$0D+#$0A;
      LogFile.write(pchar(s)^,length(s));
    end;

    j:= 0;
    load:= 0;
    del:= 0;
    tmp:= TFileListItem.create;
    count:= ResultList.Count - 1;
    stream:= nil;
    AFTPsend.ChangeWorkingDir(FTPDirBase);
    OldDir:='';;
    for I := 0 to Count do begin
      if not FAbort then begin
        tmp:= ResultList[i]as TFileListItem;
        name:= tmp.name;
        path:= ExtractFTPFilePath(name);
        path:= RemoveFTPDelimiter(path, true);
        dir:= path;
        if (tmp.UpLoad.load) or (tmp.UpLoad.overwrite) then begin
          if assigned(FProgressEvent) then FProgressEvent(load, name);
          if doload then begin
            (ResultList[i]as TFileListItem).NewList:= true;
            if WriteLog then begin
              if tmp.UpLoad.load then begin
                s:= 'l: '+tmp.name+#$0D+#$0A;
               end else begin
                s:= 'u: '+tmp.name+#$0D+#$0A;
              end;
              LogFile.write(pchar(s)^,length(s));
            end;
           end else begin
            //stream.free;
            //break;
            continue;
          end;
          inc(j);
          inc(load)
         end else if tmp.Del then begin
           if assigned(FProgressEvent) then FProgressEvent(load, name);
           name:= ExtractFTPFileName(tmp.name);
           if dodelete then begin
             (ResultList[i]as TFileListItem).NewList:= false;
             if WriteLog then begin
               s:= 'd: '+tmp.name+#$0D+#$0A;
               LogFile.write(pchar(s)^,length(s));
//               LogFile.WriteAnsiString();
             end;
             inc(j);
            end else begin
             break;
           end;
          end else begin
        end;
      end else break;
    end;
    AFTPsend.Logout;
  end;
  if assigned(LogFile) then FreeAndNil(LogFile);
  ResultList.SaveToFile(IncludeTrailingPathDelimiter(LocalDirBase) + '~~DIR~~');
end;



{ TFileListItem }

constructor TFileListItem.create;
begin
  inherited create;
  FName:= '';
  FSize:= '';
  FFTPTime:= 0;
  FLocalTime:= 0;
end;



{ TDirList }

function DoSort(Item1, Item2: Pointer): Integer;
begin
  result:= AnsiCompareStr(TFileListItem(item1).name, TFileListItem(item2).name);
end;

function TDirList.GetStrings(booleans: boolean): TStrings;
var
  i: integer;
begin
  Result:= TStringlist.Create;
  if booleans then begin
    for i := 0 to Count - 1 do begin
      Result.Add((self[i]as TFileListItem).Name+delimiter+(self[i]as TFileListItem).Time+delimiter+(self[i]as TFileListItem).Size+delimiter+(self[i]as TFileListItem).MD5+delimiter+BoolToStr((self[i]as TFileListItem).UpLoad.load, true)+BoolToStr((self[i]as TFileListItem).UpLoad.overwrite, true)+delimiter+BoolToStr((self[i]as TFileListItem).del, true));
    end
   end else begin
    for i := 0 to Count - 1 do begin
      Result.Add((self[i]as TFileListItem).Name+delimiter+(self[i]as TFileListItem).Time+delimiter+(self[i]as TFileListItem).Size+delimiter+(self[i]as TFileListItem).MD5);
    end
  end;
end;

procedure TDirList.SaveToFile(FileName: string);
var
  savelist: TStrings;
  i: integer;
begin
  savelist:= TStringList.create;
  for i := 0 to Count - 1 do begin
    if (self[i]as TFileListItem).NewList then begin
      savelist.Add((self[i]as TFileListItem).Name+delimiter+(self[i]as TFileListItem).Time+delimiter+(self[i]as TFileListItem).Size+delimiter+(self[i]as TFileListItem).MD5);
    end;
  end;
  savelist.SaveToFile(FileName);
end;

procedure TDirList.LoadFromFile(FileName: string);
var
  tmplist: TStringList;
  tmpdel: TMyStringList;
  tmpitem: TFileListItem;
  i: integer;
begin
  tmplist:= TMyStringList.Create;
  tmpdel:= TMyStringList.Create;
  tmplist.LoadFromFile(FileName);
  tmpdel.Delimiter:= delimiter;
  {$ifndef FPC}
  tmpdel.StrictDelimiter:= true;
  {$endif}
  for i := 0 to tmplist.Count - 1 do begin
    tmpitem:= TFileListItem.create;
    tmpdel.DelimitedText:= tmplist[i];
    tmpitem.Name := tmpdel[0];;
    tmpitem.Time := tmpdel[1];
    tmpitem.Size := tmpdel[2];
    tmpitem.MD5  := tmpdel[3];
    add(tmpitem);
  end;
end;

procedure TDirList.sort;
begin
  inherited Sort(@DoSort);
end;

{ TMyStringList }

procedure TMyStringList.SetDelimitedText(const AValue: string);

var i,j:integer;
    aNotFirst:boolean;
begin
// CheckSpecialChars;
 BeginUpdate;

 i:=1;
 aNotFirst:=false;

 try
  Clear;
  while i<=length(AValue) do begin
   // skip delimiter
   if aNotFirst and (i<=length(AValue)) and (AValue[i]=Delimiter) then inc(i);

   // skip spaces
   while (i<=length(AValue)) and (Ord(AValue[i])<Ord(' ')) do inc(i);

   // read next string
   if i<=length(AValue) then begin
{    if AValue[i]=QuoteChar then begin
     // next string is quoted
     j:=i+1;
     while (j<=length(AValue)) and
           ( (AValue[j]<>QuoteChar) or
             ( (j+1<=length(AValue)) and (AValue[j+1]=QuoteChar) ) ) do begin
      if (j<=length(AValue)) and (AValue[j]=QuoteChar) then inc(j,2)
                                                        else inc(j);
     end;
     // j is position of closing quote
     Add( StringReplace (Copy(AValue,i+1,j-i-1),
                         QuoteChar+QuoteChar,QuoteChar, [rfReplaceAll]));
     i:=j+1;
    end else begin     }
     // next string is not quoted
     j:=i;
     while (j<=length(AValue)) and
           (Ord(AValue[j])>=Ord(' ')) and
           (AValue[j]<>Delimiter) do inc(j);
     Add( Copy(AValue,i,j-i));
     i:=j;
 //   end;
   end else begin
    if aNotFirst then Add('');
   end;

   // skip spaces
   while (i<=length(AValue)) and (Ord(AValue[i])<Ord(' ')) do inc(i);

   aNotFirst:=true;
  end;
 finally
  EndUpdate;
 end;
end;

end.
